import re


def fnv1a32(str):
    hval = 0x811c9dc5
    prime = 0x01000193
    uint32_max = 2 ** 32
    for s in str:
        hval = hval ^ ord(s)
        hval = (hval * prime) % uint32_max
    return hval


class Experiment(object):
    def __init__(
        self,
        key: str,
        variations: list,
        weights: list = None,
        status: str = "running",
        coverage: int = 1,
        url: str = "",
        include=None,
        groups: list = None,
        force: int = None,
        hashAttribute: str = "id"
    ) -> None:
        self.key = key
        self.variations = variations
        self.weights = weights
        self.status = status
        self.coverage = coverage
        self.url = url
        self.include = include
        self.groups = groups
        self.force = force
        self.hashAttribute = hashAttribute

    def _equalWeights(self):
        return [1/len(self.variations)]*len(self.variations)

    def getWeights(self):
        weights = self.weights or self._equalWeights()

        if len(weights) != len(self.variations):
            weights = self._equalWeights()

        s = sum(weights)
        if s < 0.98 or s > 1.02:
            weights = self._equalWeights()

        if self.coverage is None:
            return weights

        coverage = self.coverage
        if coverage < 0:
            coverage = 0
        if coverage > 1:
            coverage = 1

        return list(map(lambda x: x*coverage, weights))

    def update(self, data: dict):
        weights = data.get("weights", None)
        status = data.get("status", None)
        coverage = data.get("coverage", None)
        url = data.get("url", None)
        groups = data.get("groups", None)
        force = data.get("force", None)

        if weights is not None:
            self.weights = weights
        if status is not None:
            self.status = status
        if coverage is not None:
            self.coverage = coverage
        if url is not None:
            self.url = url
        if groups is not None:
            self.groups = groups
        if force is not None:
            self.force = force


class Result(object):
    def __init__(
        self,
        variationId,
        inExperiment,
        value,
        hashAttribute,
        hashValue
    ) -> None:
        self.variationId = variationId
        self.inExperiment = inExperiment
        self.value = value
        self.hashAttribute = hashAttribute
        self.hashValue = hashValue


class GrowthBook(object):
    def __init__(
        self,
        user,
        trackingCallback=None,
        enabled=True,
        url="",
        groups={},
        overrides={},
        forcedVariations={},
        qaMode=False,
    ):
        self._user = user
        self._trackingCallback = trackingCallback
        self._enabled = enabled
        self._url = url
        self._groups = groups
        self._overrides = overrides
        self._forcedVariations = forcedVariations
        self._qaMode = qaMode

        self._tracked = {}
        self._assigned = {}
        self._subscriptions = set()

    def destroy(self):
        self._subscriptions.clear()
        self._tracked.clear()
        self._assigned.clear()
        self._trackingCallback = None
        self._forcedVariations.clear()
        self._overrides.clear()
        self._groups.clear()

    def getAllResults(self):
        return self._assigned.copy()

    def run(self, experiment: Experiment) -> Result:
        result = self._run(experiment)

        prev = self._assigned.get(experiment.key, None)
        if (not prev or prev["result"].inExperiment != result.inExperiment
                or prev["result"].variationId != result.variationId):
            self._assigned[experiment.key] = {
                "experiment": experiment,
                "result": result
            }
            for cb in self._subscriptions:
                try:
                    cb(experiment, result)
                except Exception:
                    pass

        return result

    def subscribe(self, callback):
        self._subscriptions.add(callback)
        return lambda: self._subscriptions.remove(callback)

    def _run(self, experiment: Experiment) -> Result:
        # 1. If experiment is invalid, return immediately
        if len(experiment.variations) < 2:
            return self._getResult(experiment=experiment)
        # 2. If growthbook is disabled, return immediately
        if not self._enabled:
            return self._getResult(experiment=experiment)
        # 3. If the experiment props have been overridden, merge them in
        if self._overrides.get(experiment.key, None):
            experiment.update(self._overrides[experiment.key])
        # 4. If experiment is forced via a querystring in the url
        qs = self._getQueryStringOverride(experiment.key)
        if qs is not None:
            return self._getResult(experiment=experiment, variationId=qs)
        # 5. If variation is forced in the context
        if self._forcedVariations.get(experiment.key, None) is not None:
            return self._getResult(
                experiment=experiment,
                variationId=self._forcedVariations[experiment.key]
            )
        # 6. If experiment is a draft, return immediately
        if experiment.status == "draft":
            return self._getResult(experiment=experiment)
        # 7. Get the user hash attribute and value
        hashAttribute = experiment.hashAttribute or "id"
        hashValue = self._user.get(hashAttribute, None)
        if not hashValue:
            return self._getResult(experiment=experiment)
        # 8. If experiment has an include property
        if experiment.include:
            try:
                if not experiment.include():
                    return self._getResult(experiment=experiment)
            except Exception:
                return self._getResult(experiment=experiment)
        # 9. If experiment.groups is set, make sure user is in a matching group
        if experiment.groups and len(experiment.groups):
            expGroups = self._groups or {}
            matched = False
            for group in experiment.groups:
                if expGroups[group]:
                    matched = True
            if not matched:
                return self._getResult(experiment=experiment)
        # 10. If experiment.url is set, see if it's valid
        if experiment.url:
            if not self._urlIsValid(experiment.url):
                return self._getResult(experiment=experiment)
        # 11. If experiment is forced, return immediately
        if experiment.force is not None:
            return self._getResult(
                experiment=experiment,
                variationId=experiment.force
            )
        # 12. If experiment is stopped, return immediately
        if experiment.status == "stopped":
            return self._getResult(experiment=experiment)
        # 13. If in qa mode, return immediately
        if self._qaMode:
            return self._getResult(experiment=experiment)
        # 14. Compute a hash for variation assignment
        n = (fnv1a32(hashValue + experiment.key) % 1000)/1000
        # 15. Get variation weights
        weights = experiment.getWeights()
        # 16. Loop through weights until we reach the hash value
        cumulative = 0
        assigned = -1
        for i, weight in enumerate(weights):
            cumulative += weight
            if n < cumulative and assigned < 0:
                assigned = i
        # 17. If not assigned, return immediately
        if assigned < 0:
            return self._getResult(experiment=experiment)
        result = self._getResult(
            experiment=experiment,
            variationId=assigned,
            inExperiment=True
        )
        # 18. Fire the tracking callback if set
        self._track(experiment, result)
        # 19. Return the result
        return result

    def _track(self, experiment: Experiment, result: Result):
        if not self._trackingCallback:
            return None
        key = result.hashAttribute + \
            str(result.hashValue) + experiment.key + str(result.variationId)
        if not self._tracked.get(key):
            try:
                self._trackingCallback(
                    experiment=experiment,
                    result=result
                )
                self._tracked[key] = True
            except Exception:
                pass

    def _urlIsValid(self, pattern):
        if not self._url:
            return False

        try:
            r = re.compile(pattern)
            if r.search(self._url):
                return True

            pathOnly = re.sub(
                r'^[^/]*/', '/', re.sub(r'^https?:\/\/', '', self._url))
            if r.search(pathOnly):
                return True
            return False
        except Exception:
            return True

    def _getQueryStringOverride(self, key):
        if not self._url:
            return None

        # Look for the experiment key in the querystring
        try:
            match = re.search("[?&]"+key+"=([0-9]+)($|&|#)", self._url)
            # If a match is found, return the variation, otherwise return None
            if not match:
                return None
            return int(match.group(1))
        except Exception:
            return None

    def _getResult(
        self,
        experiment,
        variationId=0,
        inExperiment=False
    ) -> Result:
        hashAttribute = experiment.hashAttribute or "id"

        if variationId < 0 or variationId > len(experiment.variations)-1:
            variationId = 0

        return Result(
            inExperiment=inExperiment,
            variationId=variationId,
            value=experiment.variations[variationId],
            hashAttribute=hashAttribute,
            hashValue=self._user.get(hashAttribute)
        )



#!/usr/bin/env python

from growthbook import GrowthBook, Experiment


def test_default_weights():
    gb = GrowthBook(user={})
    exp = Experiment(
        key="my-test",
        variations=[0, 1]
    )
    expected = [1, 0, 0, 1, 1, 1, 0, 1, 0]

    for i, v in enumerate(expected):
        gb._user = {"id": str(i+1)+""}
        assert gb.run(exp).value == v

    gb.destroy()


def test_uneven_weights():
    gb = GrowthBook(user={})
    exp = Experiment(
        key="my-test",
        variations=[0, 1],
        weights=[0.1, 0.9]
    )
    expected = [1, 1, 0, 1, 1, 1, 0, 1, 1]

    for i, v in enumerate(expected):
        gb._user = {"id": str(i+1)+""}
        assert gb.run(exp).value == v


def test_coverage():
    gb = GrowthBook(user={})
    exp = Experiment(
        key="my-test",
        variations=[0, 1],
        coverage=0.4
    )
    expected = [-1, 0, 0, -1, -1, -1, 0, -1, 1]

    for i, v in enumerate(expected):
        gb._user = {"id": str(i+1)+""}
        res = gb.run(exp)
        actual = res.variationId if res.inExperiment else -1
        assert actual == v


def test_threeWayTest():
    gb = GrowthBook(user={})

    exp = Experiment(
        key="my-test",
        variations=[0, 1, 2],
    )

    expected = [2, 0, 0, 2, 1, 2, 0, 1, 0]

    for i, v in enumerate(expected):
        gb._user = {"id": str(i+1)+""}
        assert gb.run(exp).value == v


def test_testName():
    gb = GrowthBook(user={"id": '1'})

    assert gb.run(Experiment(key="my-test", variations=[0, 1])).value == 1
    assert gb.run(Experiment(key="my-test-3", variations=[0, 1])).value == 0

    gb.destroy()


def test_missing_id():
    gb = GrowthBook(user={"id": '1'})

    exp = Experiment(
        key="my-test",
        variations=[0, 1],
    )
    assert gb.run(exp).inExperiment is True
    gb._user = {"id": ''}
    assert gb.run(exp).inExperiment is False

    gb.destroy()


def getTrackingMock(gb: GrowthBook):
    calls = []
    def track(experiment, result): return calls.append([experiment, result])
    gb._trackingCallback = track
    return lambda: calls


def test_tracking():
    gb = GrowthBook(user={"id": '1'})

    getMockedCalls = getTrackingMock(gb)

    exp1 = Experiment(
        key="my-tracked-test",
        variations=[0, 1],
    )
    exp2 = Experiment(
        key="my-other-tracked-test",
        variations=[0, 1],
    )

    res1 = gb.run(exp1)
    gb.run(exp1)
    gb.run(exp1)
    res4 = gb.run(exp2)
    gb._user = {"id": '2'}
    res5 = gb.run(exp2)

    calls = getMockedCalls()
    assert len(calls) == 3
    assert calls[0] == [exp1, res1]
    assert calls[1] == [exp2, res4]
    assert calls[2] == [exp2, res5]

    gb.destroy()


def test_handles_weird_experiment_values():
    gb = GrowthBook(user={"id": '1'})

    assert gb.run(Experiment(
        key="my-test",
        variations=[0],
    )).inExperiment is False

    assert gb.run(Experiment(
        key="my-test",
        variations=[0, 1],
        include=lambda: 1/0,
    )).inExperiment is False

    exp = Experiment(
        key="my-test",
        variations=[0, 1],
        coverage=-0.2
    )
    assert exp.getWeights() == [0, 0]

    exp.coverage = 1.5
    assert exp.getWeights() == [0.5, 0.5]

    exp.coverage = 1
    exp.weights = [0.4, 0.1]
    assert exp.getWeights() == [0.5, 0.5]

    exp.weights = [0.7, 0.6]
    assert exp.getWeights() == [0.5, 0.5]

    exp.variations = [0, 1, 2, 3]
    exp.weights = [0.4, 0.4, 0.2]
    assert exp.getWeights() == [.25, .25, .25, .25]

    res1 = gb.run(Experiment(
        key="my-test",
        variations=[0, 1],
        force=-8,
    ))

    assert res1.inExperiment is False
    assert res1.value == 0

    res2 = gb.run(Experiment(
        key="my-test",
        variations=[0, 1],
        force=25,
    ))

    assert res2.inExperiment is False
    assert res2.value == 0

    # Should fail gracefully
    gb._trackingCallback = lambda experiment, result: 1/0
    assert gb.run(Experiment(key="my-test", variations=[0, 1])).value == 1

    gb.subscribe(lambda: 1/0)
    assert gb.run(Experiment(key="my-new-test", variations=[0, 1])).value == 0

    gb.destroy()


def test_force_variation():
    gb = GrowthBook(user={"id": '6'})
    exp = Experiment(key="forced-test", variations=[0, 1])
    assert gb.run(exp).value == 0

    getMockedCalls = getTrackingMock(gb)

    gb._overrides = {
        'forced-test': {
            "force": 1,
        },
    }
    assert gb.run(exp).value == 1

    calls = getMockedCalls()
    assert len(calls) == 0

    gb.destroy()


def test_uses_overrides():
    gb = GrowthBook(
        user={"id": '1'},
        overrides={
            'my-test': {
                'coverage': 0.01,
            },
        },
    )

    assert gb.run(Experiment(
        key="my-test",
        variations=[0, 1],
    )).inExperiment is False

    gb._overrides = {
        'my-test': {
            'url': r'^\\/path',
        },
    }

    assert gb.run(Experiment(
        key="my-test",
        variations=[0, 1],
    )).inExperiment is False

    gb.destroy()


def test_filters_user_groups():
    gb = GrowthBook(
        user={"id": '123'},
        groups={
            "alpha": True,
            "beta": True,
            "internal": False,
            "qa": False,
        },
    )

    assert gb.run(Experiment(
        key="my-test",
        variations=[0, 1],
        groups=['internal', 'qa'],
    )).inExperiment is False

    assert gb.run(Experiment(
        key="my-test",
        variations=[0, 1],
        groups=['internal', 'qa', 'beta'],
    )).inExperiment is True

    assert gb.run(Experiment(
        key="my-test",
        variations=[0, 1],
    )).inExperiment is True

    gb.destroy()


def test_runs_custom_include_callback():
    gb = GrowthBook(user={"id": '1'})
    assert gb.run(Experiment(
        key="my-test",
        variations=[0, 1],
        include=lambda: False
    )).inExperiment is False

    gb.destroy()


def test_supports_custom_user_hash_keys():
    gb = GrowthBook(user={
        "id": "1",
        "company": "abc"
    })

    exp = Experiment(
        key="my-test",
        variations=[0, 1],
        hashAttribute="company"
    )

    res = gb.run(exp)

    assert res.hashAttribute == "company"
    assert res.hashValue == "abc"

    gb.destroy()


def test_experiments_disabled():
    gb = GrowthBook(user={"id": '1'}, enabled=False)

    getMockedCalls = getTrackingMock(gb)

    assert gb.run(Experiment(key="disabled-test",
                             variations=[0, 1])).inExperiment is False

    calls = getMockedCalls()
    assert len(calls) == 0

    gb.destroy()


def test_querystring_force():
    gb = GrowthBook(user={"id": '1'})
    exp = Experiment(
        key="forced-test-qs",
        variations=[0, 1],
    )
    res1 = gb.run(exp)
    assert res1.value == 0
    assert res1.inExperiment is True

    gb._url = 'http://example.com?forced-test-qs=1#someanchor'

    res2 = gb.run(exp)
    assert res2.value == 1
    assert res2.inExperiment is False

    gb.destroy()


def test_querystring_force_disabled_tracking():
    gb = GrowthBook(
        user={"id": '1'},
        url='http://example.com?forced-test-qs=1',
    )
    getMockedCalls = getTrackingMock(gb)

    exp = Experiment(
        key="forced-test-qs",
        variations=[0, 1],
    )
    gb.run(exp)

    calls = getMockedCalls()
    assert len(calls) == 0


def test_querystring_force_invalid_url():
    gb = GrowthBook(
        user={},
        url=""
    )

    gb._url = ""
    assert gb._getQueryStringOverride('my-test') is None

    gb._url = 'http://example.com'
    assert gb._getQueryStringOverride('my-test') is None

    gb._url = 'http://example.com?'
    assert gb._getQueryStringOverride('my-test') is None

    gb._url = 'http://example.com?somequery'
    assert gb._getQueryStringOverride('my-test') is None

    gb._url = 'http://example.com??&&&?#'
    assert gb._getQueryStringOverride('my-test') is None


def test_url_targeting():
    gb = GrowthBook(
        user={"id": '1'},
        url='http://example.com',
    )

    exp = Experiment(
        key="my-test",
        variations=[0, 1],
        url='^\\/post\\/[0-9]+',
    )

    res = gb.run(exp)
    assert res.inExperiment is False
    assert res.value == 0

    gb._url = 'http://example.com/post/123'
    res = gb.run(exp)
    assert res.inExperiment is True
    assert res.value == 1

    exp.url = 'http:\\/\\/example.com\\/post\\/[0-9]+'
    res = gb.run(exp)
    assert res.inExperiment is True
    assert res.value == 1

    gb.destroy()


def test_invalid_url_regex():
    gb = GrowthBook(
        user={"id": '1'},
        overrides={
            'my-test': {
                'url': '???***[)',
            },
        },
        url='http://example.com',
    )

    assert gb.run(Experiment(
        key="my-test",
        variations=[0, 1],
    )).value == 1

    gb.destroy()


def test_ignores_draft_experiments():
    gb = GrowthBook(user={"id": '1'})
    exp = Experiment(
        key="my-test",
        status='draft',
        variations=[0, 1],
    )

    res1 = gb.run(exp)
    gb._url = 'http://example.com/?my-test=1'
    res2 = gb.run(exp)

    assert res1.inExperiment is False
    assert res1.value == 0
    assert res2.inExperiment is False
    assert res2.value == 1

    gb.destroy()


def test_ignores_stopped_experiments_unless_forced():
    gb = GrowthBook(user={"id": '1'})
    expLose = Experiment(
        key="my-test",
        status='stopped',
        variations=[0, 1, 2],
    )
    expWin = Experiment(
        key="my-test",
        status='stopped',
        variations=[0, 1, 2],
        force=2,
    )

    res1 = gb.run(expLose)
    res2 = gb.run(expWin)

    assert res1.value == 0
    assert res1.inExperiment is False
    assert res2.value == 2
    assert res2.inExperiment is False

    gb.destroy()


fired = {}


def flagSubscription(experiment, result):
    fired["value"] = True


def hasFired():
    return fired.get("value", False)


def resetFiredFlag():
    fired["value"] = False


def test_destroy_removes_subscriptions():
    gb = GrowthBook(user={"id": '1'})

    resetFiredFlag()
    gb.subscribe(flagSubscription)

    gb.run(Experiment(
        key="my-test",
        variations=[0, 1],
    ))

    assert hasFired() is True

    resetFiredFlag()
    gb.destroy()

    gb.run(Experiment(
        key="my-other-test",
        variations=[0, 1],
    ))

    assert hasFired() is False

    gb.destroy()


def test_configData_experiment():
    gb = GrowthBook(user={"id": '1'})
    exp = Experiment(
        key="my-test",
        variations=[
            {
                "color": 'blue',
                "size": 'small',
            },
            {
                "color": 'green',
                "size": 'large',
            },
        ],
    )

    res1 = gb.run(exp)
    assert res1.variationId == 1
    assert res1.value == {
        "color": 'green',
        "size": 'large',
    }

    # Fallback to control config data if not in test
    exp.coverage = 0.01
    res2 = gb.run(exp)
    assert res2.inExperiment is False
    assert res2.variationId == 0
    assert res2.value == {
        "color": 'blue',
        "size": 'small',
    }

    gb.destroy()


def test_does_even_weighting():
    gb = GrowthBook(user={})
    # Full coverage
    exp = Experiment(key="my-test", variations=[0, 1])
    variations = {
        '0': 0,
        '1': 0,
        '-1': 0,
    }
    for i in range(1000):
        gb._user = {"id": str(i) + ''}
        res = gb.run(exp)
        v = res.value if res.inExperiment else -1
        variations[str(v)] += 1

    assert variations['0'] == 503

    # Reduced coverage
    exp.coverage = 0.4
    variations = {
        '0': 0,
        '1': 0,
        '-1': 0,
    }
    for i in range(1000):
        gb._user = {"id": str(i) + ''}
        res = gb.run(exp)
        v = res.value if res.inExperiment else -1
        variations[str(v)] += 1
    assert variations['0'] == 200
    assert variations['1'] == 204
    assert variations['-1'] == 596

    # 3-way
    exp.coverage = 0.6
    exp.variations = [0, 1, 2]
    variations = {
        '0': 0,
        '1': 0,
        '2': 0,
        '-1': 0,
    }
    for i in range(10000):
        gb._user = {"id": str(i) + ''}
        res = gb.run(exp)
        v = res.value if res.inExperiment else -1
        variations[str(v)] += 1
    assert variations['-1'] == 3973
    assert variations['0'] == 2044
    assert variations['1'] == 1992
    assert variations['2'] == 1991

    gb.destroy()


def test_forces_variations_from_the_client():
    gb = GrowthBook(user={"id": '1'})
    exp = Experiment(
        key="my-test",
        variations=[0, 1],
    )
    res1 = gb.run(exp)
    assert res1.inExperiment is True
    assert res1.value == 1

    gb._forcedVariations = {'my-test': 0}
    res2 = gb.run(exp)
    assert res2.inExperiment is False
    assert res2.value == 0

    gb.destroy()


def test_qa_mode():
    gb = GrowthBook(user={"id": '1'}, qaMode=True)
    exp = Experiment(
        key="my-test",
        variations=[0, 1],
    )

    res1 = gb.run(exp)
    assert res1.inExperiment is False
    assert res1.value == 0

    # Still works if explicitly forced
    gb._forcedVariations = {'my-test': 1}
    res2 = gb.run(exp)
    assert res2.inExperiment is False
    assert res2.value == 1

    # Works if the experiment itself is forced
    res3 = gb.run(Experiment(
        key="my-test-2",
        variations=[0, 1],
        force=1,
    ))

    assert res3.inExperiment is False
    assert res3.value == 1

    gb.destroy()


def test_fires_subscriptions_correctly():
    gb = GrowthBook(
        user={
            "id": '1',
        },
    )

    resetFiredFlag()
    unsubscriber = gb.subscribe(flagSubscription)

    assert hasFired() is False

    exp = Experiment(
        key="my-test",
        variations=[0, 1],
    )

    # Should fire when user is put in an experiment
    gb.run(exp)
    assert hasFired() is True

    # Does not fire if nothing has changed
    resetFiredFlag()
    gb.run(exp)
    assert hasFired() is False

    # Does not fire after unsubscribed
    unsubscriber()
    gb.run(Experiment(
        key="other-test",
        variations=[0, 1],
    ))

    assert hasFired() is False

    gb.destroy()


def test_stores_assigned_variations_in_the_user():
    gb = GrowthBook(
        user={
            "id": '1',
        },
    )

    gb.run(Experiment(key="my-test", variations=[0, 1]))
    gb.run(Experiment(key="my-test-3", variations=[0, 1]))

    assigned = gb.getAllResults()
    assignedArr = []

    for e in assigned:
        assignedArr.append({
            "key": e,
            "variation": assigned[e]["result"].variationId
        })

    assert len(assignedArr) == 2
    assert assignedArr[0]["key"] == 'my-test'
    assert assignedArr[0]["variation"] == 1
    assert assignedArr[1]["key"] == 'my-test-3'
    assert assignedArr[1]["variation"] == 0

    gb.destroy()
